"""
Unit tests for the Materials class
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys

material_test = wptherml.Materials()


def test_material_from_file():
    """tests material_from_file method using filenames for the SiO2 and TiN files,
    specifically "SiO2_ir.txt" using tabulated n and k at lambda=636 nm of 6.359999999E-7 1.45693 0.00000
    and "TiN_ellipsometry_data.txt" using tabulated n and k at lambda=1106 nm of 1.106906906906907e-06 2.175019337515494 5.175973473259225
    """

    _expected_sio2_n = 1.45693
    _expected_sio2_k = 0.00000

    _expected_tin_n = 2.175019337515494
    _expected_tin_k = 5.175973473259225

    # instance for SiO2 test
    material_test._create_test_multilayer(central_wavelength=636e-9)
    material_test.material_from_file(1, "SiO2_ir.txt")

    _result_sio2_n = np.real(material_test._refractive_index_array[1, 1])
    _result_sio2_k = np.imag(material_test._refractive_index_array[1, 1])

    # instance for TiN test
    material_test._create_test_multilayer(central_wavelength=1.106906906906907e-06)
    # define central layer as TiN
    material_test.material_from_file(1, "TiN_ellipsometry_data.txt")

    _result_tin_n = np.real(material_test._refractive_index_array[1, 1])
    _result_tin_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(_result_sio2_n, _expected_sio2_n, 1e-3)
    assert np.isclose(_result_sio2_k, _expected_sio2_k, 1e-3)
    assert np.isclose(_result_tin_n, _expected_tin_n, 1e-3)
    assert np.isclose(_result_tin_k, _expected_tin_k, 1e-3)


def test_material_sio2():
    """tests material_sio2 method using tabulated n and k at lambda=636 nm
    6.359999999E-7 1.45693 0.00000"""

    expected_n = 1.45693
    expected_k = 0.00000

    # create test multilayer that has 3 layers and wavelength array centered at 636 nm
    material_test._create_test_multilayer(central_wavelength=636e-9)
    # define central layer as siO2
    material_test.material_SiO2(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)

def test_material_sio2_udm():
    """tests material_sio2_udm method using tabulated n and k at en = 0.01207 eV:
       1.207814e-02	2.046290e+00	3.283908e-02
    """

    expected_n = 2.046290e+00
    expected_k = 3.283908e-02

    central_energy_ev = 1.207814e-02
    central_wavelength_nm = 1.2398e+3 / central_energy_ev 
    central_wavelength_m = central_wavelength_nm * 1e-9

    # create test multilayer that has 3 layers and wavelength array centered at 0.01207 eV
    material_test._create_test_multilayer(central_wavelength=central_wavelength_m)
    # define central layer as siO2
    material_test.material_SiO2_UDM(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)

def test_material_si3n4():
    """tests material_si3n4 method using tabulated n and k at lambda = 3.738 microns
       0.000003738	1.913747161	0
    """

    expected_n = 1.913747161
    expected_k = 0.0

    central_wavelength_m = 0.000003738

    # create test multilayer that has 3 layers and wavelength array centered at 3.3738 microns
    material_test._create_test_multilayer(central_wavelength=central_wavelength_m)
    # define central layer as si3n4
    material_test.material_Si3N4(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)

def test_material_zro2():
    """tests material_zro2 method using tabulated n and k at lambda = 695.2 nm 
       6.952E-07	2.144155168	0
    """

    expected_n = 2.144155168
    expected_k = 0.0

    central_wavelength_m = 6.952E-07

    # create test multilayer that has 3 layers and wavelength array centered at 695.2 nm
    material_test._create_test_multilayer(central_wavelength=central_wavelength_m)
    # define central layer as Zr02
    material_test.material_ZrO2(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)

def test_material_h2o():
    """tests material_H2O method using the fact that n = 1.33 + 0j for all lambda"""

    expected_n = 1.33
    expected_k = 0.00000

    # create test multilayer that has 3 layers and wavelength array centered at 636 nm
    material_test._create_test_multilayer(central_wavelength=636e-9)
    # define central layer as H20
    material_test.material_H2O(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_tio2():
    """tests material_tio2 method using tabulated n and k at lambda=664 nm
    6.639999999E-7 2.377021563 6.79e-10"""

    expected_n = 2.377021563
    expected_k = 6.79e-10

    # create test multilayer that has 3 layers and wavelength array centered at 664 nm
    material_test._create_test_multilayer(central_wavelength=664e-9)
    # define central layer as tiO2
    material_test.material_TiO2(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_ta2o5():
    """Dictionaries from material_Ta2O5"""
    data1 = {
        "file": "data/Ta2O5_Rodriguez.txt",
        "lower_wavelength": 2.9494e-08,
        "upper_wavelength": 1.5143e-06,
        "test_wavelength": 3.6899e-08,
        "test_n": 8.6165e-01,
        "test_k": 2.9300e-01,
    }
    data2 = {
        "file": "data/Ta2O5_Bright.txt",
        "lower_wavelength": 5.0000e-07,
        "upper_wavelength": 1.0000e-03,
        "test_wavelength": 2.5907e-06,
        "test_n": 2.0246e00,
        "test_k": 7.5989e-03,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as Ta2O5 using data1
    material_test.material_Ta2O5(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as Ta2O5 using data2
    material_test.material_Ta2O5(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_tin():
    """tests material_TiN method using tabulated n and k at lambda=1106 nm
    1.106906906906907e-06 2.175019337515494 5.175973473259225"""

    expected_n = 2.175019337515494
    expected_k = 5.175973473259225

    # create test multilayer that has 3 layers and wavelength array centered at 1106 nm
    material_test._create_test_multilayer(central_wavelength=1.106906906906907e-06)
    # define central layer as TiN
    material_test.material_TiN(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_al():
    """tests material_Al method using tabulated n and k at lambda=206.64 nm
    2.06640E-07	1.26770E-01	2.35630E+00"""

    expected_n = 1.26770e-01
    expected_k = 2.35630e00

    # create test multilayer that has 3 layers and wavelength array centered at 206.64 nm
    material_test._create_test_multilayer(central_wavelength=2.06640e-07)

    # define central layer as Al
    material_test.material_Al(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_hfo2():
    """tests material_HfO2 method using tabulated n and k at lambda=1082.00 nm
    1.082000E-06 1.880787E+00 0.000000E+00"""

    expected_n = 1.880586
    expected_k = 0.000000

    # create test multilayer that has 3 layers and wavelength array centered at 1082.0 nm
    material_test._create_test_multilayer(central_wavelength=1.082000e-06)

    # define central layer as HfO2
    material_test.material_HfO2(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_au():
    """Dictionaries from material_Au"""
    data1 = {
        "file": "data/Au_JC_RI_f.txt",
        "lower_wavelength": 2e-07,
        "upper_wavelength": 1.00025e-06,
        "test_wavelength": 5.00188e-07,
        "test_n": 0.962208410850276,
        "test_k": 1.8695066263351445,
    }
    data2 = {
        "file": "data/Au_IR.txt",
        "lower_wavelength": 3.000000e-07,
        "upper_wavelength": 2.493000e-05,
        "test_wavelength": 5.024000e-06,
        "test_n": 3.031000e00,
        "test_k": 3.447000e01,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as au using data1
    material_test.material_Au(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as au using data2
    material_test.material_Au(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_pt():
    """tests material_Pt method using tabulated n and k at lambda=610 nm
    6.1096e-06 5.4685e+00 2.4477e+01"""

    expected_n = 5.4685e00
    expected_k = 2.4477e01
    # create test multilayer that has 3 layers and wavelength array centered at 610 nm
    material_test._create_test_multilayer(central_wavelength=6.1096e-06)
    # define central layer as Pt
    material_test.material_Pt(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_al2o3():
    """tests material_al2o3 method using tabulated n and k at lambda=500 nm
    5.00E-07 1.74007 0"""

    expected_n = 1.74007
    expected_k = 0

    # create test multilayer that has 3 layers and wavelength array centered at 500 nm
    material_test._create_test_multilayer(central_wavelength=5.00e-07)
    # define central layer as Al2O3
    material_test.material_Al2O3(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)

def test_material_al2o3_udm():
    """tests material_al2o3_udm method using tabulated n and k at en = 0.01455 eV:
       1.455459e-02	2.498235e+00	6.013373e-03
    """

    expected_n = 2.498235e+00
    expected_k = 6.013373e-03

    central_energy_ev = 1.455459e-02
    central_wavelength_nm = 1.2398e+3 / central_energy_ev 
    central_wavelength_m = central_wavelength_nm * 1e-9

    # create test multilayer that has 3 layers and wavelength array centered at 636 nm
    material_test._create_test_multilayer(central_wavelength=central_wavelength_m)
    # define central layer as Al203
    material_test.material_Al2O3_UDM(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_polystyrene():
    """tests material_Polystyrene method using tabulated n and k at lambda=500 nm
    0.0000005	1.60021	6.11E-07"""

    expected_n = 1.60021
    expected_k = 6.11e-07

    # create test multilayer that has 3 layers and wavelength array centered at 500 nm
    material_test._create_test_multilayer(central_wavelength=0.0000005)
    # define central layer as polystyrene
    material_test.material_polystyrene(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_rh():
    """tests material_Rh method using tabulated n and k at lambda=564 nm
    5.636E-07	2	5.11"""

    expected_n = 2
    expected_k = 5.11
    # create test multilayer that has 3 layers and wavelength array centered at 564 nm
    material_test._create_test_multilayer(central_wavelength=5.636e-07)
    # define central layer as Rh
    material_test.material_Rh(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_ru():
    """tests material_Ru method using tabulated n and k at lambda=750 nm
    0.0000007508  5.1101514677  4.1107371518"""

    expected_n = 5.1101514677
    expected_k = 4.1107371518
    # create test multilayer that has 3 layers and wavelength array centered at 750 nm
    material_test._create_test_multilayer(central_wavelength=0.0000007508)
    # define central layer as Ru
    material_test.material_Ru(1)

    result_n = np.real(material_test._refractive_index_array[1, 1])
    result_k = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n, expected_n, 1e-3)
    assert np.isclose(result_k, expected_k, 1e-3)


def test_material_aln():
    """Dictionaries from material_AlN"""
    data1 = {
        "file": "data/AlN_Pastrnak.txt",
        "lower_wavelength": 0.000000220,
        "upper_wavelength": 0.000005000,
        "test_wavelength": 0.000001082,
        "test_n": 2.129298842,
        "test_k": 0.000000000,
    }
    data2 = {
        "file": "data/AlN_Kischkat.txt",
        "lower_wavelength": 1.53846e-06,
        "upper_wavelength": 1.42857e-05,
        "test_wavelength": 6.26566e-06,
        "test_n": 1.87617,
        "test_k": 0.00264,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as AlN using data1
    material_test.material_AlN(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])
    print(result_n_1, expected_n_1)
    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as AlN using data2
    material_test.material_AlN(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_w():
    """Dictionaries from material_W"""
    data1 = {
        "file": "data/W_Rakic.txt",
        "lower_wavelength": 1.53846e-06,
        "upper_wavelength": 1.2398e-05,
        "test_wavelength": 1.3620e-06,
        "test_n": 3.1024e00,
        "test_k": 4.3555e00,
    }

    data2 = {
        "file": "data/W_Ordal.txt",
        "lower_wavelength": 6.67000e-07,
        "upper_wavelength": 2.00000e-04,
        "test_wavelength": 2.22000e-05,
        "test_n": 27.5991880,
        "test_k": 83.5173670,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as W using data1
    material_test.material_W(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as W using data2
    material_test.material_W(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_si():
    """Dictionaries from material_Si"""
    data1 = {
        "file": "data/Si_Schinke.txt",
        "lower_wavelength": 0.000000250,
        "upper_wavelength": 0.000001450,
        "test_wavelength": 0.000000790,
        "test_n": 3.677000000,
        "test_k": 0.005688800,
    }

    data2 = {
        "file": "data/Si_Shkondin.txt",
        "lower_wavelength": 2.00000e-06,
        "upper_wavelength": 2.00000e-04,
        "test_wavelength": 0.00001052000,
        "test_n": 3.46969768200,
        "test_k": 0.00008077950,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as Si using data1
    material_test.material_Si(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as Si using data2
    material_test.material_Si(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_re():
    """Dictionaries from material_Re"""
    data1 = {
        "file": "data/Re_Windt.txt",
        "lower_wavelength": 2.36e-09,
        "upper_wavelength": 1.2157e-07,
        "test_wavelength": 5.391e-08,
        "test_n": 0.786,
        "test_k": 0.723,
    }

    data2 = {
        "file": "data/Re_Palik.txt",
        "lower_wavelength": 0.0000004000,
        "upper_wavelength": 0.0000060000,
        "test_wavelength": 0.0000007508,
        "test_n": 3.3403575532,
        "test_k": 2.7994102694,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as Re using data1
    material_test.material_Re(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as Re using data2
    material_test.material_Re(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_ag():
    """Dictionaries from material_Ag"""
    data1 = {
        "file": "data/Ag_JC.txt",
        "lower_wavelength": 1.87900e-07,
        "upper_wavelength": 1.93700e-06,
        "test_wavelength": 3.97400e-07,
        "test_n": 5.00000e-02,
        "test_k": 2.07000e00,
    }

    data2 = {
        "file": "data/Ag_Yang.txt",
        "lower_wavelength": 2.70000e-07,
        "upper_wavelength": 2.49200e-05,
        "test_wavelength": 2.08400e-06,
        "test_n": 4.61000e-01,
        "test_k": 1.47300e01,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as Ag using data1
    material_test.material_Ag(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as Ag using data2
    material_test.material_Ag(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_material_pb():
    """Dictionaries from material_Pb"""
    data1 = {
        "file": "data/Pb_Werner.txt",
        "lower_wavelength": 1.758600000e-08,
        "upper_wavelength": 2.479684000e-06,
        "test_wavelength": 3.262700000e-08,
        "test_n": 9.903000000e-01,
        "test_k": 2.206000000e-01,
    }

    data2 = {
        "file": "data/Pb_Ordal.txt",
        "lower_wavelength": 0.00000066700000,
        "upper_wavelength": 0.00066700000000,
        "test_wavelength": 0.00000286000000,
        "test_n": 3.940000000,
        "test_k": 15.600000000,
    }

    expected_n_1 = data1["test_n"]
    expected_k_1 = data1["test_k"]
    wavelength_1 = data1["test_wavelength"]

    expected_n_2 = data2["test_n"]
    expected_k_2 = data2["test_k"]
    wavelength_2 = data2["test_wavelength"]

    # create test multilayer for data1
    material_test._create_test_multilayer(central_wavelength=wavelength_1)
    # define central layer as Pb using data1
    material_test.material_Pb(1)

    result_n_1 = np.real(material_test._refractive_index_array[1, 1])
    result_k_1 = np.imag(material_test._refractive_index_array[1, 1])

    # update test multilayer for data2
    material_test._create_test_multilayer(central_wavelength=wavelength_2)
    # define central layer as Pb using data2
    material_test.material_Pb(1)

    result_n_2 = np.real(material_test._refractive_index_array[1, 1])
    result_k_2 = np.imag(material_test._refractive_index_array[1, 1])

    assert np.isclose(result_n_1, expected_n_1, 1e-3)
    assert np.isclose(result_k_1, expected_k_1, 1e-3)
    assert np.isclose(result_n_2, expected_n_2, 1e-3)
    assert np.isclose(result_k_2, expected_k_2, 1e-3)


def test_read_AM():

    # create test multilayer for a wavelength array centered at 615 nm
    material_test._create_test_multilayer(central_wavelength=615e-9)

    # Now store the AM1.5 data centered at 615 nm to an array
    _AM_15_data = material_test._read_AM()

    # this is the expected value of the AM1.5 spectrum at 615 nm
    _expected_value = 1325400000

    # test to see if the expected value is close to the read value
    assert np.isclose(_AM_15_data[1], _expected_value, 1e-3)


def test_read_atmospheric_transmissivity():

    # create test multilayer for a wavelength array centered at 7.1034 microns
    material_test._create_test_multilayer(central_wavelength=7.1034e-6)

    # Now store the AM1.5 data centered at 615 nm to an array
    _atmospheric_transmissivity = material_test._read_Atmospheric_Transmissivity()

    # this is the expected value of the AM1.5 spectrum at 615 nm
    _expected_value = 0.561289

    # test to see if the expected value is close to the read value
    assert np.isclose(_atmospheric_transmissivity[1], _expected_value, 1e-3)




    class _qubo_thickness_and_alloy_optimization:
        def __init__(self, outer_instance, optimization_dict: dict, fom_func):
            """
            pass a test_args dicticionary like
            test_args = {
                "wavelength_list": [300e-9, 6000e-9, 1000],
                "Material_List": ["Air", "Al2O3", "SiO2", "TiO2", "SiO2", "Al2O3", "W", "Air"],
                "Thickness_List": [0, 20e-9, 255e-9, 150e-9, 255e-9, 10e-9, 900e-9, 0],
                "temperature": 1700,
                "therml": True
            }

            pass another dictionary to specify which parameters to optimize
            thickness_adjustment_dict is a dictionary of indices of the materials_list to adjust thickness along with a scaling factor to explore a range of thicknesses


            explanation:
            optimization_dict = {
                "thickness_adjustment_dict": {indice_to_adjust_thickness: scaling_factor_of_original_thickness},
                "thickness_bit_length": {indice_to_adjust_thickness: number_of_bits_to_represent_thickness},
                "alloy_dict": {indice_of_alloy: [alloy_component_number_1(same as alloy indice), alloy_componenet_2]},
                "alloy_bit_length": {indice_of_alloy: number_of_bits_to represent alloy volume fraction}
            }

            example:
            optimization_dict = {
                "thickness_adjustment_dict": {1: 2},
                "thickness_bit_length": {1: 6},
                "alloy_dict": {1: [1,6]},
                "alloy_bit_length": {1: 6}
            }
            """

            # instance of outer class Opt Driver object
            self.outer_instance = outer_instance
            self.fom_func = fom_func

            self.test_args = self.outer_instance.args
            self.stack_height = len(self.test_args["Material_List"])
            self.optimization_dict = optimization_dict

            # store a copy of original test_args
            self.og_test_args = copy.deepcopy(self.outer_instance.args)
            self.og_test = copy.deepcopy(self.outer_instance)

            # get and store which indices of materials list to adjust thicknesses for
            self.thickness_adjustment_dict = optimization_dict[
                "thickness_adjustment_dict"
            ]
            # also store how many bits to represent each
            self.thickness_bit_lengths = optimization_dict["thickness_bit_length"]
            self.thickness_list = self.test_args["Thickness_List"]

            # need to get and store refractive indexes for these alloys for maxwell garnett
            # get and store indice of place where alloy will be and what the alloy make up will be
            self.alloy_dict = optimization_dict["alloy_dict"]
            # also need to store how many bits represent the alloy
            self.alloy_bit_lengths = optimization_dict["alloy_bit_length"]

            self.num_bits = 0
            # calculate total number of bits needed
            for key in self.thickness_bit_lengths:
                self.num_bits += self.thickness_bit_lengths[key]
            for key in self.alloy_bit_lengths:
                self.num_bits += self.alloy_bit_lengths[key]

            # breaking up bitstring by indice say for indice 1 you want to optimize alloy and thickness
            # it will be first bits for alloy then next bits for thickness and continue by indice
            # create a dictionary to describe how bitstring will be split up for update_multilayer function
            # of form {1: {'A': 6, 'T': 6}}
            self.bit_split_dict = {}
            for i in range(0, self.stack_height):
                if i in self.alloy_bit_lengths or i in self.thickness_bit_lengths:
                    self.bit_split_dict[i] = {}
                    if i in self.alloy_bit_lengths:
                        self.bit_split_dict[i]["A"] = self.alloy_bit_lengths[i]
                    if i in self.thickness_bit_lengths:
                        self.bit_split_dict[i]["T"] = self.thickness_bit_lengths[i]
            # print(self.bit_split_dict)

            # create a dictionary to store original ranges for the optimizaiton
            # of the form {'A': [{1: [0, 1]}], 'T': [{1: [0, 4e-08]}, {5: [0, 2e-08]}]}
            self.ranges = {"A": [], "T": []}
            for key in self.bit_split_dict:
                # key in this case is an indice
                if "A" in self.bit_split_dict[key]:
                    self.ranges["A"].append({key: [0, 1]})
                if "T" in self.bit_split_dict[key]:
                    self.ranges["T"].append(
                        {
                            key: [
                                0,
                                self.thickness_adjustment_dict[key]
                                * self.thickness_list[key],
                            ]
                        }
                    )

            print(self.ranges)

        def random_binary(self, length, num):
            # generate a 2d array of random binary string
            # specify length of binary strings as length and number you need as num
            arr = []
            for i in range(0, num):
                arr.append(list(np.random.choice([0, 1], size=(length,))))
            return arr

        def binary_to_int(self, bin):
            # convert binary to an integer
            num = int("".join(str(x) for x in bin), 2)
            return num

        def MaxwellGarnett(self, ri_1, ri_2, fraction):
            """a function that will compute the alloy refractive
            index between material_1 and material_2 using
            Maxwell-Garnett theory."""
            # define _eps_d as ri_1 ** 2
            _eps_d = ri_1 * ri_1

            # define _eps_m as ri_2 * ri_2
            _eps_m = ri_2 * ri_2

            # numerator of the Maxwell-Garnett model
            _numerator = _eps_d * (
                2 * fraction * (_eps_m - _eps_d) + _eps_m + 2 * _eps_d
            )
            # denominator of the Maxwell-Garnett model
            _denominator = 2 * _eps_d + _eps_m + fraction * (_eps_d - _eps_m)

            # _numerator / _denominator is epsilon_effective, and se want n_eff = sqrt(epsilon_eff)
            return np.sqrt(_numerator / _denominator)

        def learning_loop(
            self,
            num_iterations=1,
            num_to_train=400,
            plot_train_data=True,
            reduction_factor=2,
            n_epochs=60,
            l2_lambda=0.0001,
            l1_lambda=0,
            K=5,
            LR=0.05,
        ):
            # generate set of random bitstrings to use
            bitstrings = self.random_binary(self.num_bits, num_to_train)

            # construct a data frame with the relavent information from the random bitstrings
            """
            store data like

            bitstrings | alloy_indice_1 | thickness_indice_1 | FOM

            0001001   |    10e-9           | 0.2            | FOM
            
            """
            df = self.update_multilayer(bitstrings)

            for i in range(0, num_iterations):
                N = len((df.loc[:, "bitstrings"]).tolist()[0])
                K = K
                fm = self.outer_instance._FMTrain(N, K)
                fm.train(
                    np.array(df.loc[:, "bitstrings"]).tolist(),
                    np.array(df.loc[:, "FOM"]).reshape(-1, 1),
                    split_size=0.8,
                    batch_size=10000,
                    n_epochs=n_epochs,
                    minmaxscale=True,
                    LR=LR,
                    l2_lambda=l2_lambda,
                    l1_lambda=l1_lambda,
                    opt="LBFGS",
                )
                best_sample, samples_list = fm.get_qubo_hamiltonian_minimas(100)

                qubo_result_df = self.convert_qubo_result_to_df(best_sample.sample)
                samples_list_df = self.convert_qubo_results_to_df(samples_list)

                # add new ones to df from qubo minima
                df = pd.concat([df, samples_list_df], ignore_index=True, sort=False)

                # want to weight the lower values higher so that model learns lower regions better
                # duplicate bottom 15% of data, the 15% of data with lowest FOM, and add back to training data
                if i % 200 == 0 or i == 0:
                    df_copy = copy.deepcopy(df.sort_values(by=["FOM"]))

                    bottom_25 = int(df_copy.shape[0] * 0.15)
                    df_copy = pd.DataFrame(df_copy.iloc[0:bottom_25, :])

                    df = pd.concat([df, df_copy])

                # redo ranges every 400 iterations
                if i != 0 and i % 400 == 0:
                    qubo_result_df = df[df.FOM == df.FOM.min()]
                    qubo_result_df = qubo_result_df.reset_index()

                    # reduce ranges to
                    for j in range(0, len(self.thickness_list)):
                        # for the alloy portion'
                        for l in range(0, len(self.ranges["A"])):
                            if j in self.ranges["A"][l]:
                                center_of_new_range = qubo_result_df.loc[
                                    0, str("alloy_indice_" + str(j))
                                ]

                                new_range = (
                                    self.ranges["A"][l][j][1]
                                    - self.ranges["A"][l][j][0]
                                ) / reduction_factor

                                print(min(1, 2))
                                print(max(1, 2))
                                self.ranges["A"][l][j][0] = min(
                                    max(0, (center_of_new_range - (new_range / 2))), 1
                                )
                                self.ranges["A"][l][j][1] = min(
                                    max(0, (center_of_new_range + (new_range / 2))), 1
                                )

                        # for the thickness portion
                        for l in range(0, len(self.ranges["T"])):
                            if j in self.ranges["T"][l]:
                                center_of_new_range = qubo_result_df.loc[
                                    0, str("thickness_indice_" + str(j))
                                ]
                                new_range = (
                                    self.ranges["T"][l][j][1]
                                    - self.ranges["T"][l][j][0]
                                ) / reduction_factor

                                self.ranges["T"][l][j][0] = min(
                                    max(0, (center_of_new_range - (new_range / 2))), 1
                                )
                                self.ranges["T"][l][j][1] = min(
                                    max(0, (center_of_new_range + (new_range / 2))), 1
                                )

                        # remake data after resetting ranges

                    bitstrings = self.random_binary(self.num_bits, num_to_train)

                    # construct a data frame with the relavent information from the random bitstrings
                    """
                    store data like

                    bitstrings | alloy_indice_1 | thickness_indice_1 | FOM

                    0001001   |    10e-9           | 0.2            | FOM
                    
                    """
                    df = self.update_multilayer(bitstrings)

            return qubo_result_df.loc[:, "FOM"]

        def update_multilayer(self, bitstrings):
            """function to compute the FOM from bitstring

            Arguments
            ---------
            bistrings: dataframe with a lot of random bitstrings stored as lists

            Returns
            -------
            dataframe with FOM for that bitstring and the values represented by the bitstring
            """

            # creaating arrangement of dataframe
            columns = ["bitstrings"]

            for key in self.bit_split_dict:
                if "A" in self.bit_split_dict[key]:
                    columns.append("alloy_indice_" + str(key))

                if "T" in self.bit_split_dict[key]:
                    columns.append("thickness_indice_" + str(key))

            columns.append("FOM")

            df = pd.DataFrame(columns=columns)

            for i in range(0, len(bitstrings)):
                # store the bitstrikng in the bit strings column

                new_row_for_df = []

                current_bit = 0

                new_row_for_df.append(bitstrings[i])

                # loop through indices for optimization
                for indice in self.bit_split_dict:
                    # A or T or both for that indice
                    for key2 in self.bit_split_dict[indice]:
                        # bitstring for current split
                        bitstring_seg = bitstrings[i][
                            current_bit : current_bit
                            + self.bit_split_dict[indice][key2]
                        ]

                        current_bit = current_bit + self.bit_split_dict[indice][key2]

                        if key2 == "A":
                            # loop through the x array
                            # w = (x[0] * 2 ** 3 + x[1] * 2 ** 2 + x[2] * 2 ** 1 + x[3] * 2 ** 0) / 15
                            x = np.flip(bitstring_seg)
                            length = len(x)
                            max_x = 2 ** len(x)
                            fw = 0

                            for j in range(length):
                                fw += 2**j * x[j]

                            for k in range(0, len(self.ranges["A"])):
                                if indice in self.ranges["A"][k]:
                                    range_x = self.ranges["A"][k][indice]

                            fw = range_x[0] + fw / max_x * (range_x[1] - range_x[0])

                            # add to df
                            new_row_for_df.append(fw)

                            _n1 = np.copy(
                                self.og_test._refractive_index_array[
                                    :, self.alloy_dict[indice][0]
                                ]
                            )
                            _n2 = np.copy(
                                self.og_test._refractive_index_array[
                                    :, self.alloy_dict[indice][1]
                                ]
                            )

                            n_eff = self.MaxwellGarnett(_n1, _n2, fw)
                            self.outer_instance._refractive_index_array[
                                :, self.alloy_dict[indice][0]
                            ] = n_eff

                        if key2 == "T":
                            y = np.flip(bitstring_seg)
                            length_y = len(y)
                            max_y = 2 ** len(y)
                            val = 0

                            scaling_factor = self.thickness_adjustment_dict[indice]
                            hundred_percent = max_y / scaling_factor

                            for j in range(length_y):
                                val += 2**j * y[j]

                            percent = (val) / hundred_percent

                            for k in range(0, len(self.ranges["T"])):
                                if indice in self.ranges["T"][k]:
                                    range_y = self.ranges["T"][k][indice]

                            self.test_args["Thickness_List"][indice] = range_y[0] + (
                                self.og_test_args["Thickness_List"][indice] * percent
                            ) * (range_y[1] - range_y[0]) / (
                                self.og_test_args["Thickness_List"][indice]
                                * scaling_factor
                            )
                            self.outer_instance.thickness_array[indice] = range_y[0] + (
                                self.og_test_args["Thickness_List"][indice] * percent
                            ) * (range_y[1] - range_y[0]) / (
                                self.og_test_args["Thickness_List"][indice]
                                * scaling_factor
                            )

                            new_row_for_df.append(
                                self.test_args["Thickness_List"][indice]
                            )

                fom = self.fom_func(self.outer_instance)
                new_row_for_df.append(fom)
                df.loc[len(df.index)] = new_row_for_df

            return df

        def convert_qubo_result_to_df(self, best_sample):
            print("best_sample: ", best_sample)
            new_dict = {}
            binary_list = []

            for key in best_sample:
                temp_key = int(key[1:])
                temp_value = best_sample[key]

                new_dict[temp_key] = temp_value

            best_sample = new_dict

            myKeys = list(best_sample.keys())
            myKeys.sort()
            sorted_dict = {i: best_sample[i] for i in myKeys}

            for key in sorted_dict:
                binary_list.append(sorted_dict[key])

            minima_df = self.update_multilayer([np.array(binary_list)])

            print(minima_df)

            return minima_df

        def convert_qubo_results_to_df(self, samples_list):
            samples = []
            for m in samples_list:
                q = m.sample

                new_dict = {}
                binary_list = []

                for key in q:
                    temp_key = int(key[1:])
                    temp_value = q[key]

                    new_dict[temp_key] = temp_value

                q = new_dict

                myKeys = list(q.keys())
                myKeys.sort()
                sorted_dict = {i: q[i] for i in myKeys}

                for key in sorted_dict:
                    binary_list.append(sorted_dict[key])

                samples.append(binary_list)

            samples = [list(tupl) for tupl in {tuple(item) for item in samples}]

            minima_df = self.update_multilayer(samples)

            # print(minima_df)

            return minima_df

    class _qubo_combinatorial_structure_optimization:
        def __init__(self, outer_instance, optimization_dict, fom_func):
            # instance of outer class Opt Driver object
            self.outer_instance = outer_instance

            self.optimization_dict = optimization_dict

            self.fom_func = fom_func

            # optimizaiton dictionary with form of
            """
            for the lists of materials, the number of materials must be a power of 2, so 2, 4, 8
            layers also need to be in order no gaps unless you go to the next line so you can't have (1,2,3,5) you would need 5 on its own line below
            {
                (1,2,3,4): ("SiO2", "HfO2", "AlN", "TiO2"),
                (5,6,7,8): ("SiO2", "HfO2", "TiO2", "Al2O3")
            }
            """

            # need to parse this dictionary and construct a list of methods that hold the materials
            # to be able to modify cool_ml object
            # list of list of methods, new list for each set of layers to modify

            self.material_methods = []

            self.num_bits = 0

            for key in optimization_dict:
                materials_methods_inner = []

                materials = list(optimization_dict[key])

                # store number of bits that are needed to represent the material
                for i in range(0, len(key)):
                    self.num_bits += int(np.log((len(self.optimization_dict[key])), 2))

                # need to store methods to switch materials in structure and evaluate new structures
                for material in materials:
                    if material == "SiO2":
                        materials_methods_inner.append(
                            self.outer_instance.material_SiO2
                        )
                    elif material == "HfO2":
                        materials_methods_inner.append(
                            self.outer_instance.material_HfO2
                        )
                    elif material == "AlN":
                        materials_methods_inner.append(self.outer_instance.material_AlN)
                    elif material == "Al2O3":
                        materials_methods_inner.append(
                            self.outer_instance.material_Al2O3
                        )
                    elif material == "TiO2":
                        materials_methods_inner.append(
                            self.outer_instance.material_TiO2
                        )
                    elif material == "TiN":
                        materials_methods_inner.append(self.outer_instance.material_TiN)
                    elif material == "ZrO2":
                        materials_methods_inner.append(
                            self.outer_instance.material_ZrO2
                        )
                    elif material == "Si3N4":
                        materials_methods_inner.append(
                            self.outer_instance.material_Si3N4
                        )

                self.material_methods.append(materials_methods_inner)

            self.materials = []

            # store a list with strings of material names so that bitstrings can be mapped back to structures
            for key in optimization_dict:
                materials_inner = []

                materials = list(optimization_dict[key])

                for material in materials:
                    if material == "SiO2":
                        materials_inner.append("SiO2")
                    elif material == "HfO2":
                        materials_inner.append("HfO2")
                    elif material == "AlN":
                        materials_inner.append("AlN")
                    elif material == "Al2O3":
                        materials_inner.append("Al2O3")
                    elif material == "TiO2":
                        materials_inner.append("TiO2")
                    elif material == "TiN":
                        materials_inner.append("TiN")
                    elif material == "ZrO2":
                        materials_inner.append("ZrO2")
                    elif material == "Si3N4":
                        materials_inner.append("Si3N4")

                self.materials.append(materials_inner)

        def random_binary(self, length, num):
            # generate a 2d array of random binary string
            # specify length of binary strings as length and number you need as num
            arr = []
            for i in range(0, num):
                arr.append(list(np.random.choice([0, 1], size=(length,))))
            return arr

        def learning_loop(
            self,
            num_iterations=10,
            num_to_train=100,
            n_epochs=60,
            l2_lambda=0.0001,
            l1_lambda=0,
            K=5,
            LR=0.05,
        ):
            # generate set of random bitstrings to use
            bitstrings = self.random_binary(self.num_bits, num_to_train)

            # construct a data frame with the relavent information from the random bitstrings
            """
            store data like

            bitstrings |       materials_list               | FOM

            0001001    | ['SiO2', 'AlN', 'TiO2', 'Al2O3']   | FOM
            
            """

            # going to store all data generated in dataframe
            df = pd.DataFrame(columns=["bitstrings", "materials_list", "FOM"])

            # evaluate bitstrings to get figure of merit, FOM to create training data
            for j in range(0, len(bitstrings)):
                bitstring = bitstrings[j]
                data_from_bitstring = self.update_multilayer(bitstring)

                data_from_bitstring.insert(0, bitstring)

                df.loc[len(df.index)] = data_from_bitstring

            # loop where factorization machine is trained, minima is predicted, evaluated, and added to training data
            for i in range(0, num_iterations):
                N = len((df.loc[:, "bitstrings"]).tolist()[0])
                K = K
                fm = self.outer_instance._FMTrain(N, K)
                fm.train(
                    np.array(df.loc[:, "bitstrings"]).tolist(),
                    np.array(df.loc[:, "FOM"]).reshape(-1, 1),
                    split_size=0.8,
                    batch_size=10000,
                    n_epochs=n_epochs,
                    minmaxscale=True,
                    LR=LR,
                    l2_lambda=l2_lambda,
                    l1_lambda=l1_lambda,
                    opt="LBFGS",
                )
                best_sample = fm.get_qubo_hamiltonian_minima()

                qubo_result_list = self.convert_qubo_result_to_df(best_sample.sample)

                print(qubo_result_list)

                df.loc[len(df.index)] = qubo_result_list

        def update_multilayer(self, x):
            """function to assign a multilayer based on an input bitstring x

            Arguments
            ---------
            x : bitstring
                bitstring representation of the structure

            Returns
            -------
            net_cooling_power : float
                the figure of merit, also stored as an attribule
            """

            materials_list = []

            current_bit = 0

            i = 0
            for key in self.optimization_dict:
                bit_length = int(np.log((len(self.optimization_dict[key])), 2))

                for layer in key:
                    bits = x[current_bit : current_bit + bit_length]
                    bits = "".join(str(bit) for bit in bits)
                    num = int(bits, 2)

                    self.material_methods[i][num](layer)

                    materials_list.append(self.materials[i][num])

                    current_bit = current_bit + bit_length

                i += 1

            fom = self.fom_func(self.outer_instance)

            return [materials_list, fom]

        def convert_qubo_result_to_df(self, best_sample):
            """
            take qubo hamiltonian minima and convert result back to multilayer and place date in dataframe
            """

            new_dict = {}
            binary_list = []

            for key in best_sample:
                temp_key = int(key[1:])
                temp_value = best_sample[key]

                new_dict[temp_key] = temp_value

            best_sample = new_dict

            myKeys = list(best_sample.keys())
            myKeys.sort()
            sorted_dict = {i: best_sample[i] for i in myKeys}

            for key in sorted_dict:
                binary_list.append(sorted_dict[key])

            minima_list = self.update_multilayer(binary_list)

            minima_list.insert(0, binary_list)

            return minima_list

    class _FMTrain:
        class _TorchFM(nn.Module):
            def __init__(self, n=None, k=None):
                super().__init__()
                self.V = nn.Parameter(
                    torch.FloatTensor(n, k).uniform_(-0.1, 0.1), requires_grad=True
                )

                self.lin = nn.Linear(n, 1)

                self.n = n
                self.k = k

            def forward(self, x):
                out_1 = torch.matmul(x, self.V).pow(2).sum(1, keepdim=True)  # S_1^2
                out_2 = torch.matmul(x.pow(2), self.V.pow(2)).sum(
                    1, keepdim=True
                )  # S_2

                out_inter = 0.5 * (out_1 - out_2)

                out_lin = self.lin(x)

                out = out_inter + out_lin

                return out

            def return_params(self):
                V = self.V.detach().cpu().numpy().copy()

                return self.lin, V

        def __init__(self, n, k, verbose=False):
            # Define the model
            self.model = self._TorchFM(n, k)
            self.n = n
            self.k = k
            self.verbose = verbose

        def train(
            self,
            X,
            y,
            split_size=0.7,
            batch_size=1,
            n_epochs=1,
            minmaxscale=True,
            standardscale=False,
            opt="LBFGS",
            LR=0.01,
            l2_lambda=0.001,
            l1_lambda=0,
        ):
            """
            trains the factorization machine model, using data X and y

            """

            # train-test split for model evaluation
            (
                self.X_train_raw,
                self.X_test_raw,
                self.y_train_raw,
                self.y_test_raw,
            ) = train_test_split(X, y, train_size=split_size, shuffle=True)

            self.standardscale = standardscale
            self.minmaxscale = minmaxscale

            X_train = self.X_train_raw
            X_test = self.X_test_raw

            if self.standardscale or self.minmaxscale:
                # Standardizing data
                if self.standardscale:
                    self.scaler = StandardScaler()
                else:
                    self.scaler = MinMaxScaler((-1, 1))
                self.scaler.fit(self.y_train_raw)
                self.y_train = self.scaler.transform(self.y_train_raw)
                self.y_test = self.scaler.transform(self.y_test_raw)

            else:
                self.y_train = self.y_train_raw
                self.y_test = self.y_test_raw

            # Convert to 2D PyTorch tensors
            X_train = torch.tensor(X_train, dtype=torch.float32)
            y_train = torch.tensor(self.y_train, dtype=torch.float32).reshape(-1, 1)
            X_test = torch.tensor(X_test, dtype=torch.float32)
            y_test = torch.tensor(self.y_test, dtype=torch.float32).reshape(-1, 1)

            # loss function and optimizer
            loss_fn = nn.MSELoss()  # mean square error
            # optimizer = optim.NAdam(model.parameters(), lr = .03)

            if opt == "LBFGS":
                optimizer = optim.LBFGS(self.model.parameters(), lr=LR)
            elif opt == "ADAM":
                optimizer = optim.Adam(self.model.parameters(), lr=LR)
            elif opt == "NADAM":
                optimizer = optim.NAdam(self.model.parameters(), lr=LR)
            elif opt == "SGD":
                optimizer = optim.SGD(self.model.parameters(), lr=LR)
            else:
                optimizer = optim.Adam(self.model.parameters(), lr=LR)

            batch_start = torch.arange(0, len(X_train), batch_size)

            # Hold the best model
            self.best_mse = np.inf  # init to infinity
            self.best_weights = None
            history = []
            history_train = []

            for epoch in range(n_epochs):
                running_loss = 0
                self.model.train()
                with tqdm.tqdm(
                    batch_start, unit="batch", mininterval=0, disable=True
                ) as bar:
                    if self.verbose:
                        bar.set_description(f"Epoch {epoch}")
                    for start in bar:
                        # take a batch
                        X_batch = torch.autograd.Variable(
                            X_train[start : start + batch_size], requires_grad=True
                        )
                        y_batch = torch.autograd.Variable(
                            y_train[start : start + batch_size], requires_grad=True
                        )

                        # loss closure function
                        def closure():
                            # Zero gradients
                            optimizer.zero_grad()
                            # Forward pass
                            y_pred = self.model(X_batch)
                            # Compute loss
                            loss = loss_fn(y_pred, y_batch)

                            # l2 regularization
                            l2_norm = sum(
                                p.pow(2.0).sum() for p in self.model.parameters()
                            )

                            l1_norm = sum(
                                torch.abs(p.pow(1.0)).sum()
                                for p in self.model.parameters()
                            )

                            loss = loss + (l2_lambda * l2_norm) + (l1_lambda * l1_norm)

                            # Backward pass
                            loss.backward()
                            return loss

                        # Update weights
                        optimizer.step(closure)
                        # Update the running loss
                        loss = closure()
                        running_loss += loss.item()
                        # print progress
                        if self.verbose:
                            bar.set_postfix(mse=float(loss))
                # evaluate accuracy at end of each epoch
                self.model.eval()
                y_pred_train = self.model(X_train)
                mse_train = loss_fn(y_pred_train, y_train)
                mse_train = float(mse_train)
                y_pred = self.model(X_test)
                mse = loss_fn(y_pred, y_test)
                mse = float(mse)
                if self.verbose:
                    print(
                        "epoch-",
                        epoch,
                        "   mse------",
                        mse,
                        "   mse_train------",
                        mse_train,
                    )
                history.append(np.log(mse))
                history_train.append(np.log(mse_train))
                if mse < self.best_mse:
                    self.best_mse = mse
                    self.best_weights = copy.deepcopy(self.model.state_dict())

            # restore model and return best accuracy
            self.model.load_state_dict(self.best_weights)

            if self.verbose:
                print("MSE: %.2f" % self.best_mse)
                print("RMSE: %.2f" % np.sqrt(self.best_mse))

        def eval(self, num_eval=20):
            # for testing the inference accuracy of the model after training

            self.model.eval()
            with torch.no_grad():
                # Test out inference with 5 samples from the original test set
                for i in range(num_eval):
                    X_sample = self.X_test_raw[i : i + 1]

                    if self.standardscale or self.minmaxscale:
                        y_pred = self.scaler.inverse_transform(y_pred)

                    X_sample = torch.tensor(X_sample, dtype=torch.float32)
                    y_pred = self.model(X_sample)
                    if self.standardscale or self.minmaxscale:
                        y_pred = self.scaler.inverse_transform(y_pred)

                    print(
                        f"{self.X_test_raw[i]} -> {y_pred[0]} (expected {self.y_test[i]})"
                    )

        def get_weights(self):
            # get weights from model as list
            model_params = copy.deepcopy(self.best_weights)
            return_list = [
                np.array(model_params["lin.bias"]),
                np.array(model_params["lin.weight"]),
                np.array(model_params["V"]),
            ]
            return return_list

        def return_weights(self):
            # get weights from model
            params = self.get_weights()
            V = np.array(np.array(params[2]).tolist())
            W = np.array(list(list(np.array(params[1]))[0]), dtype=np.float64)
            bias = np.array(params[0], dtype=np.float64)[0]

            return V, W, bias

        # following 2 methods
        # https://github.com/tsudalab/fmbqm/blob/master/fmbqm/factorization_machine.py

        def triu_mask(self, input_size, F=np.ndarray):
            # Generate a square matrix with its upper trianguler elements being 1 and others 0.

            mask = np.expand_dims(np.arange(input_size), axis=0)
            return (np.transpose(mask) < mask) * 1.0

        def VtoQ(self, V, F=np.ndarray):
            """Calculate interaction strength by inner product of feature vectors."""
            print(V.shape)
            # input_size = V.shape[1]
            input_size = V.shape[0]

            # Q = np.dot(np.transpose(V), V) # (d,d)
            Q = np.dot(V, np.transpose(V))  # (d,d)
            # Q =   V @ V.T
            triu = self.triu_mask(input_size)
            # print(triu)
            # return Q * np.triu(np.ones((input_size,input_size)), 1)
            return Q * triu

        def get_qubo_hamiltonian_minima(self):
            """convert factorization machine to qubo hamiltonian and solve the model, getting one result"""

            params = copy.deepcopy(self.get_weights())

            V = np.array(params[2], dtype=np.float64)

            W = np.array(np.array(params[1])[0], dtype=np.float64)

            bias = np.array(params[0], dtype=np.float64)[0]

            Q = V @ V.T

            Q = self.VtoQ(V)

            binary_len = int(self.n)
            qubo_basis = []
            for i in range(0, binary_len):
                new_string = "x" + str(i + 1)
                qubo_basis.append(Binary(new_string))

            # define qubo Hamiltonian
            H = 0
            for i in range(0, len(Q)):
                for j in range(0, len(Q)):
                    H += Q[i][j] * qubo_basis[i] * qubo_basis[j]

            for i in range(0, len(W)):
                H += W[i] * qubo_basis[i]

            # H+= float(bias)

            # if self.verbose: print(H)
            model = H.compile()
            # if self.verbose: print(model)
            bqm = model.to_bqm()
            # if self.verbose: print(bqm)

            # solve the model
            sa = neal.SimulatedAnnealingSampler()

            sampleset = sa.sample(bqm, num_sweeps=100, num_reads=10)

            # if self.verbose: print(sampleset)
            decoded_samples = model.decode_sampleset(sampleset)
            best_sample = min(decoded_samples, key=lambda x: x.energy)

            return best_sample

        def get_qubo_hamiltonian_minimas(self, num_samples):
            """convert factorization machine to qubo hamiltonian and solve the model, getting multiple results"""

            params = copy.deepcopy(self.get_weights())

            V = np.array(params[2], dtype=np.float64)

            W = np.array(np.array(params[1])[0], dtype=np.float64)

            bias = np.array(params[0], dtype=np.float64)[0]

            Q = V @ V.T

            Q = self.VtoQ(V)

            binary_len = int(self.n)
            qubo_basis = []
            for i in range(0, binary_len):
                new_string = "x" + str(i + 1)
                qubo_basis.append(Binary(new_string))

            # define qubo Hamiltonian
            H = 0
            for i in range(0, len(Q)):
                for j in range(0, len(Q)):
                    H += Q[i][j] * qubo_basis[i] * qubo_basis[j]

            for i in range(0, len(W)):
                H += W[i] * qubo_basis[i]

            # H+= float(bias)

            # if self.verbose: print(H)
            model = H.compile()
            # if self.verbose: print(model)
            bqm = model.to_bqm()
            # if self.verbose: print(bqm)

            # solve the model

            sa = neal.SimulatedAnnealingSampler()

            samples_list = []

            for i in range(0, num_samples):
                sampleset = sa.sample(bqm, num_sweeps=1000, num_reads=10)
                # print(sampleset.record['sample'])
                for l in list(sampleset.record["sample"]):
                    samples_list.append(list(l))

            samples_list = [
                list(tupl) for tupl in {tuple(item) for item in samples_list}
            ]

            decoded_samples = model.decode_sampleset(sampleset)
            best_sample = min(decoded_samples, key=lambda x: x.energy)

            return best_sample, decoded_samples

        def predict(self, X):
            self.model.eval()
            with torch.no_grad():
                X = torch.tensor(X, dtype=torch.float32)
                predictions = self.model(X)

                if self.standardscale or self.minmaxscale:
                    predictions = self.scaler.inverse_transform(predictions)

            return predictions



import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
import os
from scipy import constants

path_and_file = os.path.realpath(__file__)
path = path_and_file[:-12]


class Materials:
    """Compute the absorption, scattering, and extinction spectra of a sphere using Mie theory"""

    def _create_test_multilayer(self, central_wavelength):
        """
        Simple method to create a 3-entry array of wavelengths as follows:
        [central_wavelength-1e-9 m, central_wavelength m, central_wavelength+1e-9 m]
        and dummy _refractive_index_array that can be filled in
        with actual materials at the wavelength arrays.
        This is simply meant to enable unit testing for desired wavelengths of the
        various materials methods
        """
        self.wavelength_array = np.array(
            [central_wavelength - 1e-9, central_wavelength, central_wavelength + 1e-9]
        )
        self.number_of_wavelengths = 3
        self.number_of_layers = 3
        self._refractive_index_array = np.reshape(
            np.tile(np.array([1 + 0j, 1 + 0j, 1 + 0j]), self.number_of_wavelengths),
            (self.number_of_wavelengths, self.number_of_layers),
        )

    def _find_unique_ri_file_data(self, wl_array):
        """
        A simple method to eliminate redundant or numerically indistinguishable
        data from refractive index files.

        Arguments
        ---------
            wl_array : numpy array of floats
                the array of wavelengths read from a refractive index file

        Returns
        -------
            unique_index_array : array of ints
                the array of the indices of the unique elements of the
                refractive index file

        """

        wl_val = wl_array[0]
        unique_index_array = [0]

        for i in range(1, len(wl_array)):
            new_wl_val = wl_array[i]
            # skip over redundant values
            if new_wl_val <= wl_val:
                wl_val = new_wl_val
            else:
                unique_index_array.append(i)
                wl_val = new_wl_val

        return unique_index_array

    def material_H2O(self, layer_number):
        """defines the refractive index layer of layer_number to be water
        assuming static refractive index of n = 1.33 + 0j
        """
        self._refractive_index_array[:, layer_number] = (
            np.ones(len(self.wavelength_array), dtype=complex) * 1.33
        )

    def insert_layer(self, layer_number):
        """insert an air layer between layer_number-1 and layer_number
        e.g. if you have a structure that is Air/SiO2/HfO2/Ag/Air
        and you issue insert_layer(1), the new structure will be
        Air/Air/SiO2/HfO2/Ag/Air
        if you issue insert_layer(2), the new structure will be
        Air/SiO2/Air/HfO2/Ag/Air

        """
        _nwl = len(self._refractive_index_array[:, 0])
        _nl = len(self._refractive_index_array[0, :])
        _temp_ri_array = np.copy(self._refractive_index_array)
        _new_ri_array = np.zeros((_nwl, _nl + 1), dtype=complex)
        _new_air_layer = np.ones(_nwl, dtype=complex) * 1.0
        _new_ri_array[:, :layer_number] = _temp_ri_array[:, :layer_number]
        _new_ri_array[:, layer_number] = _new_air_layer
        _new_ri_array[:, layer_number + 1 :] = _temp_ri_array[:, layer_number:]
        self._refractive_index_array = np.copy(_new_ri_array)

    def material_Air(self, layer_number):
        """defines the refractive index layer of layer_number to be air
        assuming static refractive index of n = 1.0 + 0j
        """
        self._refractive_index_array[:, layer_number] = (
            np.ones(len(self.wavelength_array), dtype=complex) * 1.0
        )

    def material_from_file(self, layer_number, file_name):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be
               from a text file with name "file_name" where the text file is ordered:
               column 1: wavelength in meters, increasing order
               column 2: real part of refractive index corresponding to wavelengths in col 1
               column 3: imaginary part of refractive index corresponding to wavelengths in col 1
               the file is expected to be in the directory $wpthermldir/wptherml/data
               where $wpthermldir is the full path to the directory where you have wptherml installed

            Arguments
            ----------
            layer_number : int
                specifies the layer of the stack that will be modelled as SiO2

            file_name : str
                the full file-name of the data file containing your refractive index information

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_from_file(1, "SiO2_ir.txt") -> layer 1 will be SiO2 from the SiO2_ir
            data set good from visible to 50 microns (0.21-50)
            """
            # get path to the sio2 data file
            file_path = path + "data/" + file_name
            # now read SiO2 data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            # now read Ag data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index

            # sometimes there are duplicate wavelength, n, and k entries
            # in a data set; we want only the unique elements
            idx = self._find_unique_ri_file_data(file_data[:, 0])

            n_spline = InterpolatedUnivariateSpline(
                file_data[idx, 0], file_data[idx, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[idx, 0], file_data[idx, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_2D_HOIP(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be 2D hybrid organic-inorganic
               perovskites

            Reference
            ---------
            Song et al. "Determination of dielectric functions and exciton oscillator
            strength of two-dimensional hybrid perovskites", ACS Materials Lett. 2021, 3, 1, 148-159

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as SiO2

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None
            """
            # get path to the sio2 data file
            file_path = path + "data/2D_HOIP.txt"
            # now read SiO2 data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_SiO2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be SiO2

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as SiO2

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_SiO2(1, wavelength_range="visible") -> layer 1 will be SiO2 from the SiO2_ir
            data set good from visible to 50 microns (0.21-50)
            """
            # get path to the sio2 data file
            file_path = path + "data/SiO2_ir.txt"
            # now read SiO2 data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_TiO2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be TiO2

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as TiO2

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_TiO2(1, wavelength_range="visible") -> layer 1 will be TiO2 from the Siefke data set good from visible to 125.123 microns (0.12-125.123)
            """
            # get path to the TiO2 data file
            file_path = path + "data/TiO2_Siefke.txt"
            # now read TiO2 data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Ta2O5(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Ta2O5

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Ta2O5

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Ta2O5(1, wavelength_range="visible") -> layer 1 will be Ta2O5 from the Rodriguez data set good from visible to 1.5 microns (0.02-1.5)
            >>> material_Ta2O5(2, wavelength_range="ir") -> layer 2 will be Ta2O5 from the Bright data set good until 1000 microns (0.5-1000)
            """

            # dictionary specific to Ta2O5 with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Ta2O5_Rodriguez.txt",
                "lower_wavelength": 2.9494e-08,
                "upper_wavelength": 1.5143e-06,
            }
            data2 = {
                "file": "data/Ta2O5_Bright.txt",
                "lower_wavelength": 5.0000e-07,
                "upper_wavelength": 1.0000e-03,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Ta2O5_Rodriguez.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Ta2O5_Bright.txt"

            # now read Ta2O5 data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_TiN(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Tin

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Tin

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Tin(1, wavelength_range="visible") -> layer 1 will be Tin from the Tin_ellipsometry data set good from visible to 7 microns (0.4-7)
            """
            # get path to the tin data file
            file_path = path + "data/TiN_ellipsometry_data.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_static_refractive_index(self, layer_number, refractive_index):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * refractive_index
            )

    def material_Al(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Al

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Al

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Al(1, wavelength_range="visible") -> layer 1 will be Al from the Al_Rakic data set good from visible to 200 microns (0.00012399-200)
            """
            # get path to the Al data file
            file_path = path + "data/Al_Rakic.txt"
            # now read Al data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Pt(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Pt

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Pt

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Pt(1, wavelength_range="visible") -> layer 1 will be Pt from the Pt_Rakic data set good from visible to 12.398 microns (0.24797-12.398)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the Platinum data file
            file_path = path + "data/Pt_Rakic.txt"
            # now read Platinum data into a numpy array
            file_data = np.loadtxt(file_path)
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )
            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_HfO2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be HfO2

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as HfO2

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_HfO2(1, wavelength_range="visible") -> layer 1 will be HfO2 from the HfO2_Al-Kuhaili data set good from visible to 2 microns (0.2-2)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the HfO2 data file
            file_path = path + "data/HfO2_Al-Kuhaili.txt"

            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )
            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Au(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Au

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Au
            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Au(1, wavelength_range="visible") -> layer 1 will be Au from the JC_RI_f data set  from 0.2 to 1.000025 microns
            >>> material_Au(2, wavelength_range="ir") -> layer 2 will be Au from the Au_IR data set from 0.3 to 24.93 microns
            """

            # dictionary specific to Au with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Au_JC_RI_f.txt",
                "lower_wavelength": 2e-07,
                "upper_wavelength": 1.00025e-06,
            }
            data2 = {
                "file": "data/Au_IR.txt",
                "lower_wavelength": 3.000000e-07,
                "upper_wavelength": 2.493000e-05,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Au_JC_RI_f.txt.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Au_IR.txt"

            # now read Au data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Rh(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Rh

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Rh

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Rh(1, wavelength_range="visible") -> layer 1 will be Rh from the Rh_Weaver data set good from visible to 12.4 microns (0.2-12.4)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the Rh data file
            file_path = path + "data/Rh_Weaver.txt"
            # now read Rh data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )
            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Al2O3(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Al2O3

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Al2O3

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Al2O3(1, wavelength_range="visible") -> layer 1 will be Al2O3 from the Al2O3_ri.txt data set good from visible to 2.5 microns (0.4-2.5)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the Al2O3 data file
            file_path = path + "data/Al2O3_ri.txt"
            # now read Au data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )
            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Ru(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Ru

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Ru

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Ru(1, wavelength_range="visible") -> layer 1 will be Ru from the Ru.txt data set good from visible to 6.0 microns (0.4-6.0)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the Ru data file
            file_path = path + "data/Ru.txt"
            # now read Ru data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_polystyrene(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be polystyrene

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as polystyrene

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_polystyrene(1, wavelength_range="visible") -> layer 1 will be polystyrene from the polystyrene.txt data set good from visible to 19.942 microns (0.4-19.942)
            """
            self._refractive_index_array[:, layer_number] = (
                np.ones(len(self.wavelength_array), dtype=complex) * 2.4
            )
            # get path to the polystyrene data file
            file_path = path + "data/Polystyrene.txt"
            # now read Au data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )
            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_AlN(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be AlN

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as AlN

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_AlN(1, wavelength_range="visible") -> layer 1 will be AlN from the AlN_Pastrnak data set from 0.22 to 5.0 microns
            >>> material_AlN(2, wavelength_range="ir") -> layer 2 will be AlN from the AlN_Kischkat data set from 1.53846 to 14.2857 microns
            """

            # dictionary specific to AlN with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/AlN_Pastrnak.txt",
                "lower_wavelength": 0.22e-6,
                "upper_wavelength": 5.00e-6,
            }
            data2 = {
                "file": "data/AlN_Kischkat.txt",
                "lower_wavelength": 1.53846e-06,
                "upper_wavelength": 14.2857e-06,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]
                print("1")
            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]
                print("2")
            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/AlN_Pastrnak.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/AlN_Kischkat.txt"

            # now read AlN data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_W(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be W

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as W

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_W(1, wavelength_range="visible") -> layer 1 will be W from the W_Rakic data set good from visible to 12.398 microns (0.24797-12.398)
            >>> material_W(2, wavelength_range="ir") -> layer 2 will be W from the Ordal data set good until 200 microns (0.667-200)
            """

            # dictionary specific to W with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/W_Rakic.txt",
                "lower_wavelength": 2.4797e-07,
                "upper_wavelength": 1.2398e-05,
            }
            data2 = {
                "file": "data/W_Ordal.txt",
                "lower_wavelength": 6.67000e-07,
                "upper_wavelength": 2.00000e-04,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/W_Rakic.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/W_Ordal.txt"

            # now read W data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Si(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Si

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Si

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_si(1, wavelength_range="visible") -> layer 1 will be Si from the Schinke data set good from visible to 1.5 microns (0.25-1.45)
            >>> material_si(2, wavelength_range="ir") -> layer 2 will be Si from the Shkondin data set good until 1000 microns (2-20)
            """

            # dictionary specific to W with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Si_Schinke.txt",
                "lower_wavelength": 0.000000250,
                "upper_wavelength": 0.000001450,
            }
            data2 = {
                "file": "data/Si_Shkondin.txt",
                "lower_wavelength": 0.00000200000,
                "upper_wavelength": 0.00002000000,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Si_Schinke.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Si_Shkondin.txt"

            # now read Si data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Si3N4(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Si3N4

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Si3N4

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Si3N4(1, wavelength_range="visible") -> layer 1 will be Si3N4 from the Si3N4_Luke.txt data set good from visible to 5 microns (0.361-5.14)
            """
            # get path to the Si3N4 data file
            file_path = path + "data/Si3N4_Luke.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_ZrO2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be ZrO2

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as ZrO2

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_ZrO2(1, wavelength_range="visible") -> layer 1 will be ZrO2 from the ZrO2_Wood.txt data set good from visible to 5.5 microns (0.31-5.504)
            """
            # get path to the Zr02 data file
            file_path = path + "data/ZrO2_Wood.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_SiO2_UDM(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be SiO2 using a universal dispersion model (UDM)

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as SiO2 udm

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_SiO2_UDM(1, wavelength_range="visible") -> layer 1 will be SiO2 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Si02 data file
            file_path = path + "data/SiO2_udm.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_SiO2_UDM_v2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be SiO2 using the universal dispersion model (UDM) here: http://newad.physics.muni.cz/table-udm/LithosilQ2-SPIE9890.Enk 

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as SiO2 udm

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_SiO2_UDM_v2(1, wavelength_range="visible") -> layer 1 will be SiO2 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Si02 data file
            file_path = path + "data/SiO2_udm_v2.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_HfO2_UDM(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be HfO2 using the universal dispersion model (UDM) here: http://newad.physics.muni.cz/table-udm/HfO2-X2194-AO54_9108.Enk

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as HfO2 udm

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_HfO2_UDM(1, wavelength_range="visible") -> layer 1 will be HfO2 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Si02 data file
            file_path = path + "data/HfO2_udm.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_HfO2_UDM_v2(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be HfO2 using the universal dispersion model (UDM) here: http://newad.physics.muni.cz/table-udm/HfO2-X2194-AO54_9108.Enk
               with the modification that the imaginary part of the refractive index is set to zero!

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as HfO2 udm

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_HfO2_UDM(1, wavelength_range="visible") -> layer 1 will be "lossless" HfO2 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Si02 data file
            file_path = path + "data/HfO2_udm_no_loss.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_MgF2_UDM(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be MgF2 using the universal dispersion model (UDM) here: http://newad.physics.muni.cz/table-udm/MgF2-X2935-SPIE9628.Enk
               

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as MgF2 udm

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_HfO2_UDM(1, wavelength_range="visible") -> layer 1 will be MgF2 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Si02 data file
            file_path = path + "data/MgF2_udm.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)


    def material_Al2O3_UDM(self, layer_number):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Al203

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Al203

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Al2O3_UDM(1, wavelength_range="visible") -> layer 1 will be Al2O3 from the UDM data set good from 0.01 to 100 eV
            """
            # get path to the Al203 data file
            file_path = path + "data/Al2O3_udm.txt"
            # now read Tin data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            e_data = file_data[:, 0]
            wl_nm = 1239.84193 / e_data

            wl_si = np.flip(wl_nm) * 1e-9

            n_array = np.flip(file_data[:,1])
            k_array = np.flip(file_data[:,2])

            n_spline = InterpolatedUnivariateSpline(
                wl_si, n_array, k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                wl_si, k_array, k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Re(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Re

            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Re

            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material

            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats

            Returns
            -------
            None

            Examples
            --------
            >>> material_Re(1, wavelength_range="visible") -> layer 1 will be Re from the Re_Windt data set good from visible to 1.5 microns (0.00236-0.12157)
            >>> material_Re(2, wavelength_range="ir") -> layer 2 will be Re from the Re_Palik data set good until 6 microns (0.4-6)
            """

            # dictionary specific to W with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Re_Windt.txt",
                "lower_wavelength": 2.36e-09,
                "upper_wavelength": 1.2157e-07,
            }
            data2 = {
                "file": "data/Re_Palik.txt",
                "lower_wavelength": 0.0000004000,
                "upper_wavelength": 0.0000060000,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Re_Windt.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Re_Palik.txt"

            # now read Re data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Ag(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Ag
            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Ag
            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material
            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats
            Returns
            -------
            None
            Examples
            --------
            >>> material_Ag(1, wavelength_range="visible") -> layer 1 will be Ag from the Ag_JC data set good from 0.1879 to 1.937
            >>> material_Ag(2, wavelength_range="ir") -> layer 2 will be Ag from the Yang data set good until 24.92 microns (0.27-24.92)
            """

            # dictionary specific to W with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Ag_JC.txt",
                "lower_wavelength": 1.87900e-07,
                "upper_wavelength": 1.93700e-06,
            }
            data2 = {
                "file": "data/Ag_Yang.txt",
                "lower_wavelength": 2.70000e-07,
                "upper_wavelength": 2.49200e-05,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Ag_JC.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Ag_Yang.txt"

            # now read Ag data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index

            # sometimes there are duplicate wavelength, n, and k entries
            # in a data set; we want only the unique elements
            idx = self._find_unique_ri_file_data(file_data[:, 0])

            n_spline = InterpolatedUnivariateSpline(
                file_data[idx, 0], file_data[idx, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[idx, 0], file_data[idx, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def material_Pb(self, layer_number, wavelength_range="visible", override="true"):
        if layer_number > 0 and layer_number < (self.number_of_layers - 1):
            """defines the refractive index of layer layer_number to be Pb
            Arguments
            ----------
            layer_number : int
            specifies the layer of the stack that will be modelled as Pb
            wavelength_range (optional) : str
            specifies wavelength regime that is desired for modelling the material
            Attributes
            ----------
            _refractive_index_array : 1 x number_of_wavelengths numpy array of complex floats
            Returns
            -------
            None
            Examples
            --------
            >>> material_pb(1, wavelength_range="visible") -> layer 1 will be Pb from the Werner data set good from visible to 2.47 microns
            >>> material_pb(2, wavelength_range="ir") -> layer 2 will be Pb from the ordal data set good until 667 microns
            """

            # dictionary specific to W with wavelength range information corresponding to different
            # data sets
            data1 = {
                "file": "data/Pb_Werner.txt",
                "lower_wavelength": 1.758600000e-08,
                "upper_wavelength": 2.479684000e-06,
            }
            data2 = {
                "file": "data/Pb_Ordal.txt",
                "lower_wavelength": 0.00000066700000,
                "upper_wavelength": 0.00066700000000,
            }

            shortest_wavelength = self.wavelength_array[0]
            longest_wavelength = self.wavelength_array[self.number_of_wavelengths - 1]

            if (
                shortest_wavelength >= data1["lower_wavelength"]
                and longest_wavelength <= data1["upper_wavelength"]
            ):
                file_path = path + data1["file"]

            elif (
                shortest_wavelength >= data2["lower_wavelength"]
                and longest_wavelength <= data2["upper_wavelength"]
            ):
                file_path = path + data2["file"]

            else:
                file_path = path + data1["file"]

            if override == "false":
                # make sure the wavelength_range string is all  lowercase
                wavelength_range = wavelength_range.lower()
                if (
                    wavelength_range == "visible"
                    or wavelength_range == "short"
                    or wavelength_range == "vis"
                ):
                    file_path = path + "data/Pb_Werner.txt"

                elif wavelength_range == "ir" or wavelength_range == "long":
                    file_path = path + "data/Pb_Ordal.txt"

            # now read Pb data into a numpy array
            file_data = np.loadtxt(file_path)
            # file_path[:,0] -> wavelengths in meters
            # file_path[:,1] -> real part of the refractive index
            # file_path[:,2] -> imaginary part of the refractive index
            n_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 1], k=1
            )
            k_spline = InterpolatedUnivariateSpline(
                file_data[:, 0], file_data[:, 2], k=1
            )

            self._refractive_index_array[:, layer_number] = n_spline(
                self.wavelength_array
            ) + 1j * k_spline(self.wavelength_array)

    def _read_CIE(self):
        """Reads CIE data and stores as attributes self.cie_cr, self.cie_cg, self.cie_cb

        Arguments
        ----------
        None

        References
        ----------
        Equations 29-32 of https://github.com/FoleyLab/wptherml/blob/master/docs/Equations.pdf

        Attributes
        ----------
        _cie_cr : 1 x wavelength array of floats
            data corresponding to red cone response function in integrand of Eq. 29

        _cie_cg : 1 x wavelength array of floats
            data corresponding to the green cone response function in integrand of Eq. 30

        _cie_cb : 1 x wavelength array of floats
            data corresponding to the blue cone response function in integrand of Eq. 31

        Returns
        -------
        None
        """
        # initialize cie arrays?
        self._cie_cr = np.zeros_like(self.wavelength_array)
        self._cie_cg = np.zeros_like(self.wavelength_array)
        self._cie_cb = np.zeros_like(self.wavelength_array)
        # )
        # get path to the cie data
        file_path = path + "data/cie_cmf.txt"
        # now read Rh data into a numpy array
        file_data = np.loadtxt(file_path)
        # file_data[:,0] -> wavelengths in nm
        # file_data[:,1] -> cr response function
        # file_data[:,2] -> cg response function
        # file_data[:,3] -> cb resposne function

        _cr_spline = InterpolatedUnivariateSpline(
            file_data[:, 0] * 1e-9, file_data[:, 1], k=1
        )
        _cg_spline = InterpolatedUnivariateSpline(
            file_data[:, 0] * 1e-9, file_data[:, 2], k=1
        )
        _cb_spline = InterpolatedUnivariateSpline(
            file_data[:, 0] * 1e-9, file_data[:, 3], k=1
        )
        # values of data file at 500 nm
        expected_values = np.array([0.0049, 0.3230, 0.2720])
        spline_values = np.array(
            [_cr_spline(500e-9), _cg_spline(500e-9), _cb_spline(500e-9)]
        )
        assert np.allclose(expected_values, spline_values)
        self._cie_cr[:] = _cr_spline(self.wavelength_array)
        self._cie_cg[:] = _cg_spline(self.wavelength_array)
        self._cie_cb[:] = _cb_spline(self.wavelength_array)

    def _read_AM(self):
        """Reads AM1.5 data and returns an array of the AM1.5 data evaluated at each value of
            self.wavelength_array

        Arguments
        ----------
        None

        References
        ----------
        add

        Attributes
        ----------
        None

        Returns
        -------
        1 x number_of_wavelengths array of floats
            Data corresponding the AM1.5 spectrum evaluated at each value of self.wavelength_array

        """

        # get path to the AM data
        file_path = path + "data/scaled_AM_1_5.txt"
        # now read Rh data into a numpy array
        file_data = np.loadtxt(file_path)
        # file_data[:,0] -> wavelengths in m
        # file_data[:,1] -> solar spectrum in W / m / m^2 / sr

        _solar_spline = InterpolatedUnivariateSpline(
            file_data[:, 0], file_data[:, 1], k=1
        )

        # values of data file at 615 nm
        # 0.000000615000000       1325400000.0000000000000000000000
        _expected_value = 1325400000.0
        _spline_value = _solar_spline(615e-9)
        assert np.isclose(_expected_value, _spline_value)
        return _solar_spline(self.wavelength_array)

    def _read_Atmospheric_Transmissivity(self):
        """Reads atmospherical transmissivity data and returns
            an array of this data evaluated at each value of self.wavelength_array

        Arguments
        ----------
        None

        References
        ----------
        add

        Attributes
        ----------
        None

        Returns
        -------
        1 x number_of_wavelengths array of floats
            atmospheric transmissivity evaluated at each value of self.wavelength_array
        """

        # get path to the AM data
        file_path = path + "data/Atmospheric_transmissivity.txt"
        # now read Rh data into a numpy array
        file_data = np.loadtxt(file_path)
        # file_data[:,0] -> wavelengths in m
        # file_data[:,1] -> atmospheric transmissivity

        # get indices of unique elements
        idx = self._find_unique_ri_file_data(file_data[:, 0])

        _atrans_spline = InterpolatedUnivariateSpline(
            file_data[idx, 0], file_data[idx, 1], k=1
        )

        # values of data file at 7.1034e-06 meters (7.1034 microns) -> T = 0.561289
        _expected_value = 0.561289
        _spline_value = _atrans_spline(7.1034e-6)
        assert np.isclose(_expected_value, _spline_value)
        return _atrans_spline(self.wavelength_array)
    
    def _EQE_spectral_response(self):
        """ 
        Will compute the spectral response function using tabulated EQE values for user input thickness
            from 

            "Optical Properties and Modeling of 2D Perovskite Solar Cells",
            Bin Liu, Chan Myae Myae Soe, Constantinos C. Stoumpos, Wanyi Nie, Hsinhan Tsai, Kimin
            Lim, Aditya D. Mohite, Mercouri G. Kanatzidis, Tobin J. Marks, Kenneth D. Singer
            Advanced Materials, (34), 1. January 6, 2022 
            https://doi.org/10.1002/adma.202107211

            on EQE curves of Pb5 perovskite-based devices.

            Using formula SR = q * EQE * \lambda / (h * c)

        Attributes
        ----------
        psc_thickness : 
                        User-input thickness in nm of PSC.
        _eqe_array :
                        Table of EQE values extrapolated from graph.
        _sr_array : 
                        Calculated spectral response based on formula, wavelength, tabulated values, and constants.
        _eqe_spline :
                        Spline best fit based on wavelength array and EQE data.
        _sr_spline :
                        Spline best fit based on SR array and wavelength array.

        Returns
        -------
        None
        
        """

        # Initialize wavelength array and variable for thickness
        _wavelength_array = np.array([250, 300, 350, 400, 450, 500, 550, 600, 650, 700, 750, 800, 850]) * 1e-9

        # Check for psc values
        if self.psc_thickness_option == 110:
            _eqe_array = np.array([0, 0, 22, 46, 58, 59, 54, 40, 24, 18, 4,	0, 0])

        elif self.psc_thickness_option == 200:
            _eqe_array = np.array([0, 0, 25, 52, 72, 74, 71, 59, 39, 22, 5, 0, 0])

        elif self.psc_thickness_option == 250:
            _eqe_array = np.array([0, 0, 20, 40, 54, 59, 58, 48, 39, 32, 14, 0,	0])
        
        elif self.psc_thickness_option == 410:
            _eqe_array = np.array([0, 0, 16, 27, 33, 36, 35, 32, 31, 24, 7, 0, 0])
                    
        else:
            _eqe_array = np.array([0, 0, 25, 52, 72, 74, 71, 59, 39, 22, 5, 0, 0])

        # Transferring from percentages and SR calculation
        _eqe_array = _eqe_array * 0.01
        _sr_array = constants.e * _eqe_array * _wavelength_array / (constants.h * constants.c)

        # Spline for line of best fit based on current data
        _eqe_spline = InterpolatedUnivariateSpline(
            _wavelength_array, _eqe_array, k=1
        )

        _sr_spline = InterpolatedUnivariateSpline(
                _wavelength_array, _sr_array, k=1
            )
        
        self.perovskite_eqe = _eqe_spline(self.wavelength_array)
        self.perovskite_spectral_response = _sr_spline(self.wavelength_array)
    




"""
Unit and regression test for the wpspec package.
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys


def test_wptherml_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "wptherml" in sys.modules



"""
Unit and regression test for the wpspec package.
"""

# Import package, test suite, and other packages as needed
import wptherml
import numpy as np
import pytest
import sys


def test_compute_power_density():
    test_args = {
        "wavelength_list": [100e-9, 30000e-9, 10000],
        "material_list": ["Air", "W", "Air"],
        "thickness_list": [0, 800e-9, 0],
        "temperature": 1500,
        "therml": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)

    """will test _compute_power_density method to
    see if the power density computed by integration
    of the blackbody spectrum is close to the prediction
    by the Stefan-Boltzmann law, where the latter should be exact
    """
    assert np.isclose(test.blackbody_power_density, test.stefan_boltzmann_law, 1e-2)


def test_compute_stpv_power_density():

    # define basic structure at 1500 K
    test_args = {
        "wavelength_list": [400e-9, 7000e-9, 1000],
        "material_list": ["Air", "TiN", "Air"],
        "thickness_list": [0, 400e-9, 0],
        "temperature": 5000,
        "therml": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)
    """will test _compute_stpv_power_density method to
    see if the power density computed by integration
    of the thermal emission spectrum of TiN at 5000 K is close to 
    the value that is generated by v1.0.0 of wptherml
    """
    assert np.isclose(test.stpv_power_density, 3863945.0, 1e4)


def test_compute_stpv_gradients():
    """unit test for the computation of the gradients of stpv quantities
    including the stpv_power_density and the spectral_efficiency"""
    test_args = {
        "wavelength_list": [400e-9, 7000e-9, 1000],
        "material_list": ["Air", "SiO2", "TiN", "Air"],
        "thickness_list": [0, 20e-9, 400e-9, 0],
        "temperature": 5000,
        "therml": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)
    test.compute_stpv_gradient()

    # get analytic gradients
    _analytic_stpv_power_density_gradient = test.stpv_power_density_gradient[0]
    _analytic_stpv_spectral_efficiency_gradient = (
        test.stpv_spectral_efficiency_gradient[0]
    )

    # define a displacement in thickness of SiO2
    _delta_d_sio2 = 0.1e-9
    # take forward step and store quantities
    test.thickness_array[1] += _delta_d_sio2
    test.compute_spectrum()
    test.compute_stpv()
    _stpv_power_density_f = test.stpv_power_density
    _stpv_spectral_efficiency_f = test.stpv_spectral_efficiency

    # take backward step and store quantities
    test.thickness_array[1] -= 2 * _delta_d_sio2
    test.compute_spectrum()
    test.compute_stpv()
    _stpv_power_density_b = test.stpv_power_density
    _stpv_spectral_efficiency_b = test.stpv_spectral_efficiency

    # compute gradients by centered fininte differences
    _numeric_stpv_power_density_gradient = (
        _stpv_power_density_f - _stpv_power_density_b
    ) / (2 * _delta_d_sio2)
    _numeric_stpv_spectral_efficiency_gradient = (
        _stpv_spectral_efficiency_f - _stpv_spectral_efficiency_b
    ) / (2 * _delta_d_sio2)

    # these gradients have big values so scale them relative to the numeric gradient
    _scaled_analytic_stpv_power_density_gradient = (
        _analytic_stpv_power_density_gradient / _numeric_stpv_power_density_gradient
    )

    # test stpv_power_density_gradient
    assert np.isclose(_scaled_analytic_stpv_power_density_gradient, 1, 2e-2)
    # test stpv_spectral_efficiency_gradient
    assert np.isclose(
        _numeric_stpv_spectral_efficiency_gradient,
        _analytic_stpv_spectral_efficiency_gradient,
        1e-2,
    )


def test_compute_stpv_efficiency():
    # define basic structure at 1500 K
    test_args = {
        "wavelength_list": [400e-9, 7000e-9, 1000],
        "material_list": ["Air", "TiN", "Air"],
        "thickness_list": [0, 400e-9, 0],
        "temperature": 1700,
        "therml": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)
    """will test _compute_stpv_power_density method to
    see if the power density computed by integration
    of the thermal emission spectrum of TiN at 5000 K is close to 
    the value that is generated by v1.0.0 of wptherml
    """
    assert np.isclose(test.stpv_spectral_efficiency, 0.3786094154907156, 5e-2)


def test_compute_luminous_efficiency():
    # define basic structure at 1500 K
    test_args = {
        "wavelength_list": [400e-9, 7000e-9, 1000],
        "material_list": ["Air", "TiN", "Air"],
        "thickness_list": [0, 400e-9, 0],
        "temperature": 5000,
        "therml": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)
    """will test _compute_stpv_power_density method to
    see if the power density computed by integration
    of the thermal emission spectrum of TiN at 5000 K is close to 
    the value that is generated by v1.0.0 of wptherml
    """
    assert np.isclose(test.luminous_efficiency, 0.20350729803724107, 1e-2)


def test_compute_cooling():
    test_args = {
        "wavelength_list": [300e-9, 60000e-9, 5000],
        "material_list": ["Air", "SiO2", "Air"],
        "thickness_list": [0, 230e-9, 0],
        "temperature": 300,
        "cooling": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)
    test._refractive_index_array[:, 1] = 2.4 + 0.2j

    _expected_radiative_cooling_power = 40.44509251986298
    _expected_atmospheric_warming_power = 21.973817620650525
    _expected_solar_warming_power = 426.9132402277394

    test.compute_cooling()

    test.compute_explicit_angle_spectrum()

    assert np.isclose(
        _expected_radiative_cooling_power, test.radiative_cooling_power, 1e-5
    )
    assert np.isclose(
        _expected_atmospheric_warming_power, test.atmospheric_warming_power, 1e-5
    )
    assert np.isclose(_expected_solar_warming_power, test.solar_warming_power, 1e-5)


def test_compute_cooling_gradient():
    """FINISH THIS UNIT TEST"""

    # define basic structure at 1500 K
    test_args = {
        "wavelength_list": [300e-9, 20000e-9, 5000],
        "material_list": ["Air", "TiN", "Air"],
        "thickness_list": [0, 230e-9, 0],
        "temperature": 300,
        "cooling": True,
    }
    sf = wptherml.SpectrumFactory()
    test = sf.spectrum_factory("Tmm", test_args)

    # get analytic gradient of the absorbed solar power
    test.compute_cooling_gradient()

    # define a displacement in thickness of SiO2
    _delta_d_sio2 = 1.0e-9

    test.incident_angle = test.solar_angle
    # get forward solar power
    test.thickness_array[1] += _delta_d_sio2
    test.compute_cooling()

    _solar_power_f = test.solar_warming_power
    _emitted_power_f = test.radiative_cooling_power
    _atmospheric_power_f = test.atmospheric_warming_power

    # get backward solar power
    test.thickness_array[1] -= 2 * _delta_d_sio2
    test.compute_cooling()
    _solar_power_b = test.solar_warming_power
    _emitted_power_b = test.radiative_cooling_power
    _atmospheric_power_b = test.atmospheric_warming_power

    _numeric_solar_warming_power_gradient = (_solar_power_f - _solar_power_b) / (
        2 * _delta_d_sio2
    )
    _numeric_radiative_cooling_power_gradient = (
        _emitted_power_f - _emitted_power_b
    ) / (2 * _delta_d_sio2)
    _numeric_atmospheric_warming_power_gradient = (
        _atmospheric_power_f - _atmospheric_power_b
    ) / (2 * _delta_d_sio2)

    # normalize the gradients by the numeric gradient
    _normalized_analytic_solar_power_gradient = (
        test.solar_warming_power_gradient / _numeric_solar_warming_power_gradient
    )
    _normalized_analytic_emitted_power_gradient = (
        test.radiative_cooling_power_gradient
        / _numeric_radiative_cooling_power_gradient
    )
    _normalized_analytic_atmospheric_power_gradient = (
        test.atmospheric_warming_power_gradient
        / _numeric_atmospheric_warming_power_gradient
    )

    # if the gradients are close, the normalized analytic gradient will be close to 1
    assert np.isclose(_normalized_analytic_solar_power_gradient[0], 1.0, 1e-3)
    assert np.isclose(_normalized_analytic_emitted_power_gradient[0], 1.0, 1e-3)
    assert np.isclose(_normalized_analytic_atmospheric_power_gradient[0], 1.0, 1e-3)
    assert np.isclose(
        test.atmospheric_warming_power_gradient[0],
        _numeric_atmospheric_warming_power_gradient,
        1e-2,
    )



from .spectrum_driver import SpectrumDriver

# child classes
from .mie import MieDriver
from .em import TmmDriver
from .therml import Therml
from .optdriver import OptDriver
from .exciton import ExcitonDriver
from .spin_boson import SpinBosonDriver 


class SpectrumFactory:
    def spectrum_factory(self, spectrum_toolkit, args):
        if spectrum_toolkit == "Mie":
            return MieDriver(args)
        elif spectrum_toolkit == "Tmm":
            return TmmDriver(args)
        elif spectrum_toolkit == "Opt":
            return OptDriver(args)
        elif spectrum_toolkit == "Frenkel":
            return ExcitonDriver(args)
        elif spectrum_toolkit == "Spin-Boson":
            return SpinBosonDriver(args)
        else:
            raise TypeError("Toolkit not found")


"""class SpectrumFactory:
    _toolkits = {}

    def spectrum_factory(self, spectrum_toolkit, size):
        if spectrum_toolkit not in self._toolkits.keys():
            raise Exception('Toolkit not found.')
        cls = self._toolkits[spectrum_toolkit]
        return cls(size)

    def register(self, toolkit_name, toolkit_class):
        if not issubclass(toolkit_class, SpectrumAdapter):
            raise TypeError(f'{toolkit_class} is not a SpectrumAdapter')
        self._toolkits[toolkit_name] = toolkit_class
"""
